#!/usr/bin/env python3

# Copyright © 2023 Scott Court
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shlex
import sys
import traceback
import datetime

from isidore.libIsidore import *

# The Isidore command prompt
class IsidoreCmdline:

    _isidore = None
    _version = '0.1.3'

    # Creates a new Isidore command prompt
    # @param isidore    The underlying Isidore instance for the command prompt
    #                   to connect to.
    def __init__(self, isidore):
        self._isidore = isidore

    # Gets the Isidore Command Prompt version
    # @return       The Isidore Command Prompt version
    def getVersion(self):
        return self._version

    # Start a subprompt.
    # @param prompt     The existing arguments to display at the
    #                   prompt and prepend to read command
    #                   arguments before passing to func.
    # @param func       The function to use to process input from
    #                   the subprompt.
    def subprompt(self, prompt, func):
        line = []
        while line != ['end']:
            # Determine prompt
            if sys.stdin.isatty():
                display_prompt = ' '.join(prompt) + '> '
            else:
                display_prompt = ''

            # Read input
            try:
                line = shlex.split(input(display_prompt))
            except EOFError:
                print()
                return
            except KeyboardInterrupt:
                print('^C')
                continue
            except:
                print("Malformed command", file=sys.stderr)
                continue

            # Process input
            if line == []:
                continue
            elif line == ['end']:
                return
            elif line == ['quit']:
                exit()
            elif line[0] == '?':
                print('''\
^C          clear the current command
^D          alias for end
end         go back to the previous prompt
quit        exit''')

            func(prompt + line)

    # Start an interactive prompt
    def prompt(self):
        self.subprompt([], self.rootprompt)

    # >
    def rootprompt(self, args):

        # Parse inut
        if args[0] == '?':
            self.help(args)
        elif args[0] == 'create':
            self.create(args)
        elif args[0] == 'delete':
            self.delete(args)
        elif args[0] == 'echo':
            self.echo(args)
        elif args[0] == 'help':
            print('''\
Pst! You should really use ? to display the help message. ? will
work at every subprompt level. help only works at the root
prompt.
''')
            self.help(args)
        elif args[0] == 'host':
            self.host(args)
        elif args[0] == 'rename':
            self.rename(args)
        elif args[0] == 'show':
            self.show(args)
        elif args[0] == 'tag':
            self.tag(args)
        elif args[0] == 'version':
            self.version(args)
        else:
            print('Invalid command '+args[0]+'. Enter ? for help.', file=sys.stderr)

    # > ?
    def help(self, args):
        print('''\
?           print this help message
create      create various objects (such as hosts and tags)
delete      delete various objects (such as hosts and tags)
echo        print text back to the console
help        alias for ?
host        manipulate a host
rename      rename various objects (such as hosts and tags)
show        print various data
tag         manipulate a tag
version     display Isidore version information''')

    # > show
    def show(self, args):
        if len(args) == 1:
            self.subprompt(args, self.show)

        elif args[1] == '?':
            print('''\
?           print this help message
config      print the commmands to populate the database with the current 
            configuration
hosts       print all commissioned hosts in the database
graveyard   print all decommissioned hosts in the database
inventory   print the full Ansible inventory file
tag-groups  print all the tag groups in the database
tags        print all tags in the database''')

        elif args[1] == 'config':
            self.show_config(args[1])

        elif args[1] == 'hosts':
            self.show_hosts(args[1])

        elif args[1] == 'inventory':
            self.show_inventory(args)

        elif args[1] == 'graveyard':
            self.show_graveyard(args[1])

        elif args[1] == 'tag-groups':
            self.show_taggroups(args[1])

        elif args[1] == 'tags':
            self.show_tags(args[1])

        else:
            print('Invalid argument '+args[1]+'. Enter ? for help.', file=sys.stderr)

    # > show config
    def show_config(self, args):
        hosts = self._isidore.getHosts()
        tags = self._isidore.getTags()

        # Create Hosts
        for host in hosts:
            name = "'"+host.getHostname().replace("'", "'\"'\"'")+"'"
            print("create host "+name)
            print("host "+name+" set commissioned '"+\
                    str(host.getCommissionDate())+"'")
            print("host "+name+" set decommissioned '"+\
                    str(host.getDecommissionDate()).lower()+"'")
            if host.getDescription() == None:
                print("host "+name+" set description none")
            else:
                print("host "+name+" set description '"+\
                        str(host.getDescription()).replace("'", "'\"'\"'")+"'")
            print("host "+name+" var set $ '"+\
                    json.dumps(host.getVar()).replace("'", "'\"'\"'")+"'")
        print()

        # Create Tags
        for tag in tags:
            name = "'"+tag.getName().replace("'", "'\"'\"'")+"'"
            group = tag.getGroup()
            description = tag.getDescription()
            print("create tag %s" % (name))
            print("tag %s set group '%s'" %
                    (name,
                    'none' if group == None
                        else group.replace("'", "'\"'\"'")))
            print("tag %s set description '%s'" % 
                    (name,
                    'none' if description == None
                        else description.replace("'", "'\"'\"'")))
            print("tag "+name+" var set $ '"+\
                    json.dumps(tag.getVar()).replace("'", "'\"'\"'")+"'")
        print()

        # Assign Tags to Hosts
        for host in hosts:
            hostname = "'"+host.getHostname().replace("'", "'\"'\"'")+"'"
            for tag in host.getTags():
                print("host "+hostname+" tag add '"+\
                        str(tag.getName()).replace("'", "'\"'\"'")+"'")

    # > show graveyard
    def show_graveyard(self, args):
        for host in self._isidore.getDecommissionedHosts():
            print(host.getHostname())

    # > show hosts
    def show_hosts(self, args):
        for host in self._isidore.getCommissionedHosts():
            print(host.getHostname())

    # > show inventory
    def show_inventory(self, args):
        if len(args) == 2:
            print(self._isidore.getInventoryIni())

        elif args[2] == '?':
            print('''\
?           print this help message
ini         print the inventory in INI format
json        print the inventory in JSON format
yaml        print the inventory in YAML format''')

        elif args[2] == 'ini':
            print(self._isidore.getInventoryIni())
        elif args[2] == 'json':
            print(self._isidore.getInventoryJson())
        elif args[2] == 'yaml':
            print(self._isidore.getInventoryYaml())
        else:
            print('Invalid format '+args[2]+'. Enter ? for help.', file=sys.stderr)

    # > show tag-groups
    def show_taggroups(self, args):
        tags = {}

        # Populate the top level
        for (group, tagsstr) in self._isidore.getTagGroups():
            if group == None:
                group = 'ungrouped'
            tags[group] = list()

        for tag in self._isidore.getTags():
            group = tag.getGroup()
            if group == None:
                group = 'ungrouped'
            tags[group].append(tag.getName())

        print(yaml.dump(tags, default_flow_style=False))

    # > show tags
    def show_tags(self, args):
        for tag in self._isidore.getTags():
            print(tag.getName())

    # > create
    def create(self, args):
        if len(args) == 1:
            self.subprompt(args, self.create)
        elif args[1] == '?':
            print('''\
?           print this help message
host        create a new host
tag         create a new tag''')
        elif args[1] == 'host':
            self.create_host(args)
        elif args[1] == 'tag':
            self.create_tag(args)
        else:
            print('Invalid argument '+args[1]+'. Enter ? for help.', file=sys.stderr)

    # > create host
    def create_host(self, args):
        if len(args) == 2:
            self.subprompt(args, self.create)
        elif args[2] == '?':
            print('''\
?           print this help message
<hostname>  the hostname for the new host to create''')
        else:
            try:
                self._isidore.createHost(args[2])
            except mysql.connector.errors.IntegrityError as e:
                print('Host %s already exists' % args[2],
                        file=sys.stderr)
            except:
                print('Failed to create host '+args[2],
                        file=sys.stderr)
                print(traceback.format_exc(),
                        file=sys.stderr)

    # > create tag
    def create_tag(self, args):
        if len(args) == 2:
            self.subprompt(args, self.create)
        elif args[2] == '?':
            print('''\
?           print this help message
<name>      the name of the new tag to create''')
        else:
            try:
                self._isidore.createTag(args[2])
            except mysql.connector.errors.IntegrityError as e:
                print('Tag %s already exists' % args[2],
                        file=sys.stderr)
            except:
                print('Failed to create tag '+args[2],
                        file=sys.stderr)
                print(traceback.format_exc(),
                        file=sys.stderr)

    # > delete
    def delete(self, args):
        if len(args) == 1:
            self.subprompt(args, self.delete)
        elif args[1] == '?':
            print('''\
?           print this help message
host        delete a host
tag         delete a tag''')
        elif args[1] == 'host':
            self.delete_host(args)
        elif args[1] == 'tag':
            self.delete_tag(args)
        else:
            print('Invalid argument '+args[1]+'. Enter ? for help.', file=sys.stderr)

    # > delete host
    def delete_host(self, args):
        if len(args) == 2:
            self.subprompt(args, self.delete_host)
            return
        elif args[2] == '?':
            print('''\
?           print this help message
<hostname>  the hostname of the host to delete''')
            return

        host = self._isidore.getHost(args[2])
        if host == None:
            print('Host '+args[2]+' does not exist!')
            return

        try:
            host.delete()
            print("Host "+args[2]+" has been deleted.")
        except mysql.connector.Error as e:
            if e.errno == 1451:
                print("Cannot delete host "+host.getHostname()+": it still has tags assigned to it.", file=sys.stderr)
            else:
                print('Failed to delete host '+args[2],
                        file=sys.stderr)
                print(traceback.format_exc(),
                        file=sys.stderr)
        except:
            print('Failed to delete host '+args[2],
                    file=sys.stderr)
            print(traceback.format_exc(),
                    file=sys.stderr)

    # > delete tag
    def delete_tag(self, args):
        if len(args) == 2:
            self.subprompt(args, self.delete_tag)
            return
        elif args[2] == '?':
            print('''\
?           print this help message
<name>      the name of the tag to delete''')
            return

        tag = self._isidore.getTag(args[2])
        if tag == None:
            print('Tag '+args[2]+' does not exist!')
            return

        try:
            tag.delete()
            print("Tag "+args[2]+" has been deleted.")
        except mysql.connector.Error as e:
            if e.errno == 1451:
                print("Cannot delete tag "+tag.getName()+": it still has hosts assigned to it.", file=sys.stderr)
            else:
                print('Failed to delete tag '+args[2],
                        file=sys.stderr)
                print(traceback.format_exc(),
                        file=sys.stderr)
        except:
            print('Failed to delete tag '+args[2],
                    file=sys.stderr)
            print(traceback.format_exc(),
                    file=sys.stderr)

    # > echo
    def echo(self, args):
        if len(args) == 1:
            self.subprompt(args, self.echo)
        elif args[1] == '?':
            print('''\
?           print this help message
<text>      text to print''')
        else:
            print(' '.join(args[1:]))

    # > host
    def host(self, args):
        # Handle arg #1 (host <ARG1>)
        if len(args) == 1:
            self.subprompt(args, self.host)
            return
        elif args[1] == '?':
            print('''\
?           print this help message
<hostname>  the name of the host to edit''')
            return
        host = self._isidore.getHost(args[1])
        if host == None:
            print('Host '+args[1]+' does not exist!')
            return

        # Handle arg #2 (host foo <ARG2>)
        elif len(args) == 2:
            self.subprompt(args, self.host)
        elif args[2] == '?':
            print('''\
?           print this help message
set         modify host attributes
show        display host attributes
tag         display and modify this host's tags
var         display and modify this host's variables''')
        elif args[2] == 'set':
            self.host_set(args)
        elif args[2] == 'show':
            self.host_show(args)
        elif args[2] == 'tag':
            self.host_tag(args)
        elif args[2] == 'var':
            self.host_var(args)
        else:
            print('Invalid command '+args[2]+'. Enter ? for help.', file=sys.stderr)

    # > host <hostname> show
    def host_show(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 3:
            self.subprompt(args, self.host)
        elif args[3] == '?':
            print('''\
?           print this help message
all         print all the information about the host
commissioned    print the date the host was commissioned
description     print the host's description
decommissioned  print the date the host was decommissioned''')
        elif args[3] == 'all':
            print(yaml.dump(host.getDetails(), default_flow_style=False))
        elif args[3] == 'commissioned':
            print(host.getCommissionDate())
        elif args[3] == 'description':
            print(host.getDescription())
        elif args[3] == 'decommissioned':
            print(host.getDecommissionDate())
        else:
            print('Invalid argument '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > host <hostname> set
    def host_set(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 3:
            self.subprompt(args, self.host)
        elif args[3] == '?':
            print('''\
?           print this help message
commissioned    set the date the host was commissioned
description     set the host's description
decommissioned  set the date the host was decommissioned''')
        elif args[3] == 'commissioned':
            self.host_set_commissioned(args)
        elif args[3] == 'decommissioned':
            self.host_set_decommissioned(args)
        elif args[3] == 'description':
            if len(args) == 4:
                print("<description>   the description")
            elif args[4] == 'none':
                host.setDescription(None)
            else:
                host.setDescription(args[4])
        else:
            print('Invalid argument '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > host <hostname> set commissioned
    def host_set_commissioned(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host)
        elif args[4] == '?':
            print('''\
<date>      the commission date
now         use the current date''')
        elif args[4] == 'now':
            host.setCommissionDate(datetime.datetime.now())
        else:
            try:
                host.setCommissionDate(args[4])
            except:
                print("Failed to set commission date")

    # > host <hostname> set decommissioned
    def host_set_decommissioned(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host)
        elif args[4] == '?':
            print('''\
<date>      the decommission date
none        clear the decommission date
now         use the current date''')
        elif args[4] == 'none':
            host.setDecommissionDate(None)
        elif args[4] == 'now':
            host.setDecommissionDate(datetime.datetime.now())
        else:
            try:
                host.setDecommissionDate(args[4])
            except:
                print("Failed to set decommission date")

    # > host <hostname> tag
    def host_tag(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 3:
            self.subprompt(args, self.host)
        elif args[3] == '?':
            print('''\
?           print this help message
add         add a tag to this host
list        list the tags currently assigned to this host
list-detail display a detailed list of tags currently assigned to this host
remove      remove a tag from this host''')

        elif args[3] == 'add':
            self.host_tag_add(args)

        elif args[3] == 'list':
            for tag in host.getTags():
                print(tag.getName())

        elif args[3] == 'list-detail':
            tags = list()
            for tag in host.getTags(True):
                tags.append( {
                    'name': tag.getName(),
                    'group': tag.getGroup(),
                    'description': tag.getDescription()
                    } )
            print(yaml.dump(tags, default_flow_style=False))

        elif args[3] == 'remove':
            self.host_tag_remove(args)
        else:
            print('Invalid command '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > host <hostname> var
    def host_var(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 3:
            self.subprompt(args, self.host_var)
        elif args[3] == '?':
            print('''\
?           print this help message
append      append a value to a list variable
print       print a variable
set         set a variable
unset       unset (delete) a variable''')

        elif args[3] == 'append':
            self.host_var_append(args)

        elif args[3] == 'print':
            if len(args) == 4:
                print(yaml.dump(
                    host.getVar(),
                    default_flow_style=False,
                    sort_keys=False))
            elif args[4] == '?':
                print('''\
?           print this help message
$           print all variables
<variable>  name of the variable to print''')
            else:
                print(yaml.dump(
                    host.getVar(args[4]),
                    default_flow_style=False,
                    sort_keys=False))

        elif args[3] == 'set':
            self.host_var_set(args)

        elif args[3] == 'unset':
            self.host_var_unset(args)

        else:
            print('Invalid command '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > host <hostname> var append
    def host_var_append(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host_var_append)
        elif args[4] == '?':
            print('''\
?           print this help message
<variable>  name of the list variable to append to''')

        elif len(args) == 5:
            self.subprompt(args, self.host_var_append)

        elif args[5] == '?':
            print('''\
?           print this help message
<json>      the JSON value to append to the list''')

        else:
            try:
                host.appendVar(args[4], json.loads(args[5]))
            except json.decoder.JSONDecodeError:
                print(args[5] + '''
^-- this is not valid JSON

Strings must be double quoted. It will be necessary to either nest double
quotes inside single quotes or escape the double quotes like so:

   > host myhost var set foo '"bar"'

or

   > host myhost var set foo \\"bar\\"

''')
            except:
                print(\
'Failed to append to list variable. Is %s a valid list path?' % args[4])

    # > host <hostname> var set
    def host_var_set(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host_var_set)
        elif args[4] == '?':
            print('''\
?           print this help message
$           set/replace the entire variable tree
<variable>  name of the variable to set''')

        elif len(args) == 5:
            self.subprompt(args, self.host_var_set)

        elif args[5] == '?':
            print('''\
?           print this help message
<json>      the JSON value to set the variable to''')

        else:
            try:
                host.setVar(args[4], json.loads(args[5]))
            except json.decoder.JSONDecodeError:
                print(args[5] + '''
^-- this is not valid JSON

Strings must be double quoted. It will be necessary to either nest double
quotes inside single quotes or escape the double quotes like so:

   > host myhost var set foo '"bar"'

or

   > host myhost var set foo \\"bar\\"

''')

    # > host <hostname> var unset
    def host_var_unset(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host_var_set)
        elif args[4] == '?':
            print('''\
?           print this help message
<variable>  name of the variable to unset''')
        else:
            try:
                host.unsetVar(args[4])
            except:
                print("Failed to unset variable %s" % args[4])

    # > host <hostname> tag add
    def host_tag_add(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host)
            return
        elif args[4] == '?':
            print('''\
?           print this help message
<tag>       name of the tag to add''')
            return

        tag = self._isidore.getTag(args[4])
        if tag == None:
            print("Tag "+args[4]+" does not exist")
            return
        try:
            host.addTag(tag)
        except mysql.connector.Error as e:
            if e.errno == 1062:
                print(host.getHostname()+" already has tag "+args[4], file=sys.stderr)
        except:
            print(traceback.format_exc(),
                    file=sys.stderr)

    # > host <hostname> tag remove
    def host_tag_remove(self, args):
        host = self._isidore.getHost(args[1])
        if len(args) == 4:
            self.subprompt(args, self.host)
            return
        elif args[4] == '?':
            print('''\
?           print this help message
<tag>       name of the tag to remove''')
            return
        tag = self._isidore.getTag(args[4])
        if tag == None:
            print("Tag "+args[4]+" does not exist")
            return
        host.removeTag(tag)

    # > rename
    def rename(self, args):
        if len(args) == 1:
            self.subprompt(args, self.rename)
        elif args[1] == '?':
            print('''\
?           print this help message
host        rename a host
tag         rename a tag''')
        elif args[1] == 'host':
            self.rename_host(args)
        elif args[1] == 'tag':
            self.rename_tag(args)
        else:
            print('Invalid argument '+args[1]+'. Enter ? for help.', file=sys.stderr)

    # > rename host <old_hostname> <new_hostname>
    def rename_host(self, args):
        if len(args) == 2:
            self.subprompt(args, self.rename_host)
            return
        elif args[2] == '?':
            print('''\
?           print this help message
<hostname>  the old hostname''')
            return

        host = self._isidore.getHost(args[2])

        if host == None:
            print("Host "+args[2]+" does not exist.")
            return
        elif len(args) == 3:
            print(\
'''Rename does not allow for a subprompt for the fourth argument. You must
enter both the old and new hostnames at the same time 

Example:
    > rename host foo bar

Enter ? as any argument help.''', file=sys.stderr)
            return
        elif args[3] == '?':
            print('''\
?           print this help message
<hostname>  the new hostname''')
            return

        try:
            host.setHostname(args[3])
        except mysql.connector.Error as e:
            if e.errno == 1062:
                print('Host '+args[3]+' already exists.', file=sys.stderr)
            else:
                print(traceback.format_exc(), file=sys.stderr)
        except:
            print(traceback.format_exc(), file=sys.stderr)

    # > rename tag <old_hostname> <new_hostname>
    def rename_tag(self, args):
        if len(args) == 2:
            self.subprompt(args, self.rename_tag)
            return
        elif args[2] == '?':
            print('''\
?           print this help message
<name>      the old tag name''')
            return

        tag = self._isidore.getTag(args[2])

        if tag == None:
            print("Tag "+args[2]+" does not exist.")
            return
        elif len(args) == 3:
            print(\
'''Rename does not allow for a subprompt for the fourth argument. You must
enter both the old and new tag names at the same time 

Example:
    > rename tag foo bar

Enter ? as any argument help.''', file=sys.stderr)
            return
        elif args[3] == '?':
            print('''\
?           print this help message
<name>      the new tag name''')
            return

        try:
            tag.setName(args[3])
        except mysql.connector.Error as e:
            if e.errno == 1062:
                print('Tag '+args[3]+' already exists.', file=sys.stderr)
            else:
                print(traceback.format_exc(), file=sys.stderr)
        except:
            print(traceback.format_exc(), file=sys.stderr)

    # > tag
    def tag(self, args):
        # Handle arg #1 (tag <ARG1>)
        if len(args) == 1:
            self.subprompt(args, self.tag)
            return
        elif args[1] == '?':
            print('''\
?           print this help message
<tagname>  the name of the tag to edit''')
            return
        tag = self._isidore.getTag(args[1])
        if tag == None:
            print('Tag '+args[1]+' does not exist!')
            return

        # Handle arg #2 (tag foo <ARG2>)
        elif len(args) == 2:
            self.subprompt(args, self.tag)
        elif args[2] == '?':
            print('''\
?           print this help message
host        display and modify hosts that have this tag
set         modify tag attributes
show        display tag attributes
var         display and modify this tag's variables''')
        elif args[2] == 'host':
            self.tag_host(args)
        elif args[2] == 'set':
            self.tag_set(args)
        elif args[2] == 'show':
            self.tag_show(args)
        elif args[2] == 'var':
            self.tag_var(args)
        else:
            print('Invalid command '+args[2]+'. Enter ? for help.', file=sys.stderr)

    # > tag <tagname> host
    def tag_host(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 3:
            self.subprompt(args, self.tag)
        elif args[3] == '?':
            print('''\
?           print this help message
add         assign hosts to this tag
list        display all hosts that have this tag
remove      remove hosts from this tag''')
        elif args[3] == 'add':
            self.tag_host_add(args)
        elif args[3] == 'list':
            for host in tag.getHosts():
                print(host.getHostname())
        elif args[3] == 'remove':
            self.tag_host_remove(args)
        else:
            print('Invalid command '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > tag <tagname> host add
    def tag_host_add(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag)
            return
        elif args[4] == '?':
            print('''\
?           print this help message
<host>      name of the host to add''')
            return

        host = self._isidore.getHost(args[4])
        if host == None:
            print("Host "+args[4]+" does not exist")
            return
        try:
            host.addTag(tag)
        except mysql.connector.Error as e:
            if e.errno == 1062:
                print(host.getHostname()+" already has tag "+args[1], file=sys.stderr)
        except:
            print(traceback.format_exc(),
                    file=sys.stderr)

    # > tag <tagname> host remove
    def tag_host_remove(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag)
            return
        elif args[4] == '?':
            print('''\
?           print this help message
<host>      name of the host to remove''')
            return
        host = self._isidore.getHost(args[4])
        if host == None:
            print("Host "+args[4]+" does not have tag "+args[1])
            return
        host.removeTag(tag)

    # > tag <tagname> show
    def tag_show(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 3:
            self.subprompt(args, self.tag)
        elif args[3] == '?':
            print('''\
?           print this help message
all         print all the information about the tag
description print the tag's description
group       print the date the tag was commissioned''')
        elif args[3] == 'all':
            print(yaml.dump(tag.getDetails(), default_flow_style=False))
        elif args[3] == 'description':
            print(tag.getDescription())
        elif args[3] == 'group':
            print(tag.getGroup())
        else:
            print('Invalid argument '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > tag <tagname> set
    def tag_set(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 3:
            self.subprompt(args, self.tag)
        elif args[3] == '?':
            print('''\
?           print this help message
description     set the tag's description
group           set the tag's group''')
        elif args[3] == 'group':
            self.tag_set_group(args)
        elif args[3] == 'description':
            if len(args) == 4:
                print("<description>   the description")
            elif args[4] == 'none':
                tag.setDescription(None)
            else:
                tag.setDescription(args[4])
        else:
            print('Invalid argument '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > tag <tagname> set group
    def tag_set_group(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag)
        elif args[4] == '?':
            print('''\
<group>     the tag group
none        remove tag group''')
        elif args[4] == 'none':
            tag.setGroup(None)
        else:
            try:
                tag.setGroup(args[4])
            except:
                print("Failed to set group")

    # > tag <tagname> var
    def tag_var(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 3:
            self.subprompt(args, self.tag_var)
        elif args[3] == '?':
            print('''\
?           print this help message
append      append a value to a list variable
print       print a variable
set         set a variable
unset       unset (delete) a variable''')

        elif args[3] == 'append':
            self.tag_var_append(args)

        elif args[3] == 'print':
            if len(args) == 4:
                print(yaml.dump(
                    tag.getVar(),
                    default_flow_style=False,
                    sort_keys=False))
            elif args[4] == '?':
                print('''\
?           print this help message
$           print all variables
<variable>  name of the variable to print''')
            else:
                print(yaml.dump(
                    tag.getVar(args[4]),
                    default_flow_style=False,
                    sort_keys=False))

        elif args[3] == 'set':
            self.tag_var_set(args)

        elif args[3] == 'unset':
            self.tag_var_unset(args)

        else:
            print('Invalid command '+args[3]+'. Enter ? for help.', file=sys.stderr)

    # > tag <tagname> var append
    def tag_var_append(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag_var_append)
        elif args[4] == '?':
            print('''\
?           print this help message
<variable>  name of the list variable to append to''')

        elif len(args) == 5:
            self.subprompt(args, self.tag_var_append)

        elif args[5] == '?':
            print('''\
?           print this help message
<json>      the JSON value to append to the list''')

        else:
            try:
                tag.appendVar(args[4], json.loads(args[5]))
            except json.decoder.JSONDecodeError:
                print(args[5] + '''
^-- this is not valid JSON

Strings must be double quoted. It will be necessary to either nest double
quotes inside single quotes or escape the double quotes like so:

   > tag mytag var set foo '"bar"'

or

   > tag mytag var set foo \\"bar\\"

''')
            except:
                print(\
'Failed to append to list variable. Is %s a valid list path?' % args[4])

    # > tag <tagname> var set
    def tag_var_set(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag_var_set)
        elif args[4] == '?':
            print('''\
?           print this help message
$           set/replace the entire variable tree
<variable>  name of the variable to set''')

        elif len(args) == 5:
            self.subprompt(args, self.tag_var_set)

        elif args[5] == '?':
            print('''\
?           print this help message
<json>      the JSON value to set the variable to''')

        else:
            try:
                tag.setVar(args[4], json.loads(args[5]))
            except json.decoder.JSONDecodeError:
                print(args[5] + '''
^-- this is not valid JSON

Strings must be double quoted. It will be necessary to either nest double
quotes inside single quotes or escape the double quotes like so:

   > tag mytag var set foo '"bar"'

or

   > tag mytag var set foo \\"bar\\"

''')

    # > tag <tagname> var unset
    def tag_var_unset(self, args):
        tag = self._isidore.getTag(args[1])
        if len(args) == 4:
            self.subprompt(args, self.tag_var_set)
        elif args[4] == '?':
            print('''\
?           print this help message
<variable>  name of the variable to unset''')
        else:
            try:
                tag.unsetVar(args[4])
            except:
                print("Failed to unset variable %s" % args[4])

    # > version
    def version(self, args):
            print('Isidore Command Prompt version: '+self.getVersion())
            print('libIsidore version: '+self._isidore.getVersion())
            print('Isidore database version: '+self._isidore.getDatabaseVersion())




#!/usr/bin/env python3

# Copyright © 2023 Scott Court
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import configparser

import mysql.connector
import yaml
import json

# Represents an Isidore database instance
class Isidore:

    _conn = None
    _version = '0.1.3'

    # Connects to a MySQL database and creates a new Isidore object to interact
    # with it.
    # @param user       The MySQL username
    # @param password   The password for the MySQL user
    # @param host       The MySQL server to connect to
    # @param database   The name of the database to use
    def __init__(self, user, password, host, database):
        self._conn = mysql.connector.connect(
                user = user,
                password = password,
                host = host,
                database = database
        )

    # Loads the database credentials from a file. It then connects to the MySQL
    # database specified by the config and creates a new Isidore object to
    # interact with it.
    #
    # If no file is specified, the standard Isidore system config files will be
    # tried in the following order. Any variables set in a file that is further
    # down the list will take precedence over anything set in a file higher up
    # the list.
    # - /etc/isidore.cfg
    # - /usr/local/etc/isidore.cfg
    # - ~/.isidore.cfg
    # - ./isidore.cfg
    #
    # @param file       The path to the file to load, or None to use the system
    #                   configuration.
    @classmethod
    def fromConfigFile(cls, file=None):

        # Read config file
        config = configparser.ConfigParser()
        if file == None:
            config.read( [
                "/etc/isidore.cfg",
                "/usr/local/etc/isidore.cfg",
                os.path.expanduser("~/.isidore.cfg"),
                "isidore.cfg"
                ] )
        else:
            config.read( [ file ] )

        # Set the required variables
        user = config['database']['user']
        password = config['database']['password']
        host = config['database']['host']
        database = config['database']['database']

        # Make the MySQL connection
        return cls(user, password, host, database)

    # Creates a new host in the database
    # @param hostname           The hostname for the new host
    def createHost(self, hostname):
        cursor = self._conn.cursor()
        stmt = "INSERT INTO Host (Hostname) VALUES (%s)"
        cursor.execute(stmt, [ hostname ])
        self._conn.commit()
        cursor.close()

    # Creates a new tag in the database
    # @param name               The name for the new tag
    def createTag(self, name):
        cursor = self._conn.cursor()
        stmt = "INSERT INTO Tag (TagName) VALUES (%s)"
        cursor.execute(stmt, [ name ])
        self._conn.commit()
        cursor.close()

    # Gets all the commissioned hosts in the database
    # @return   An array containing all the commissioned hosts in
    #           the database
    def getCommissionedHosts(self):
        hosts = list()

        cursor = self._conn.cursor()
        cursor.execute('''
                SELECT
                    HostID,
                    Hostname,
                    CommissionDate,
                    DecommissionDate,
                    Description
                FROM Host
                WHERE DecommissionDate IS NULL
                ORDER BY Hostname ASC''')
        for (hostId, hostname, commissionDate, decommissionDate, description) in cursor:
            host = Host(hostId, hostname, commissionDate,
                    decommissionDate, description, self)
            hosts.append(host)
        cursor.close()

        return hosts

    # Gets the underlying Isidore database version.
    # @return       The Isidore database version
    def getDatabaseVersion(self):
        cursor = self._conn.cursor()
        cursor.execute("SELECT Value FROM Metadata WHERE KeyName = 'version'")
        row = cursor.fetchone()
        cursor.close()
        return row[0]

    # Gets all the decommissioned hosts in the database
    # @return   An array containing all the decommissioned hosts in
    #           the database
    def getDecommissionedHosts(self):
        hosts = list()

        cursor = self._conn.cursor()
        cursor.execute('''
                SELECT
                    HostID,
                    Hostname,
                    CommissionDate,
                    DecommissionDate,
                    Description
                FROM Host
                WHERE DecommissionDate IS NOT NULL
                ORDER BY Hostname ASC''')
        for (hostId, hostname, commissionDate, decommissionDate, description) in cursor:
            host = Host(hostId, hostname, commissionDate,
                    decommissionDate, description, self)
            hosts.append(host)
        cursor.close()

        return hosts

    # Gets a host in the database
    # @param hostname   The hostname of the system to get
    # @return           The Host object, or None if the host does
    #                   not exist.
    def getHost(self, hostname):
        cursor = self._conn.cursor()
        cursor.execute('''
                SELECT
                    HostID,
                    Hostname,
                    CommissionDate,
                    DecommissionDate,
                    Description
                FROM Host
                WHERE Hostname = %s''',
                [hostname])

        row = cursor.fetchone()
        if row == None:
            return None

        host = Host(row[0], row[1], row[2], row[3], row[4],
                self)
        cursor.fetchall()
        cursor.close()
        return host

    # Gets all the hosts in the database
    # @return   An array containing all the hosts in the database
    def getHosts(self):
        hosts = list()

        cursor = self._conn.cursor()
        cursor.execute('''
                SELECT
                    HostID,
                    Hostname,
                    CommissionDate,
                    DecommissionDate,
                    Description
                FROM Host
                ORDER BY Hostname ASC''')
        for (hostId, hostname, commissionDate, decommissionDate, description) in cursor:
            host = Host(hostId, hostname, commissionDate,
                    decommissionDate, description, self)
            hosts.append(host)
        cursor.close()

        return hosts

    # Builds an Ansible inventory in a dictionary representation
    # from all hosts and tags in the database
    # @return       the Ansible inventory dictionary
    #               representation 
    def getInventory(self):
        inv = dict()

        # Add all the hosts without a group header to ensure every
        # system is included, even those without any tags.
        inv['all'] = {
                'hosts': list()
        }
        for host in self.getCommissionedHosts():
            inv['all']['hosts'].append(host.getDetails())

        # Add each tag and its hosts as a group
        for tag in self.getTags(True):
            name = tag.getName()
            inv[name] = tag.getDetails()[name]

        return inv

    # Builds an Ansible INI inventory from all hosts and tags in
    # the database
    # @return       the Ansible inventory INI representation as a
    #               string
    def getInventoryIni(self):
        inv = ""

        # Add all the hosts without a group header to ensure every
        # system is included, even those without any tags.
        inv += "# All Host\n"
        for host in self.getCommissionedHosts():
            inv += host.getHostname() + "\n"
        inv += "\n"

        # Print each tag and its hosts as a group
        for tag in self.getTags(True):
            # Comment
            group = tag.getGroup()
            if group == None:
                inv += "# "+tag.getName()+\
                        " ("+str(tag.getDescription())+")\n"
            else:
                inv += "# "+tag.getGroup()+": "+tag.getName()+\
                        " ("+str(tag.getDescription())+")\n"

            # Header
            inv += "["+tag.getName()+"]\n"

            # Hosts
            for host in tag.getHosts():
                inv += host.getHostname() + "\n"
            inv += "\n"

        return inv

    # Builds an Ansible inventory in JSON format from all hosts
    # and tags in the database
    # @return       the Ansible JSON inventory as a string
    def getInventoryJson(self):
        inv = dict()

        # Add meta section
        inv['_meta'] = {
                'hostvars': {}
        }

        # Add all the hosts without a group header to ensure every
        # system is included, even those without any tags.
        inv['all'] = {
                'hosts': list()
        }
        for host in self.getCommissionedHosts():
            name = host.getHostname()
            inv['all']['hosts'].append(name)
            inv['_meta']['hostvars'][name] = host.getDetails()[name]['vars']

        # Add each tag and its hosts as a group
        for tag in self.getTags(True):
            name = tag.getName()
            inv[name] = tag.getDetails()[name]

        return json.dumps(inv)

    # Builds an Ansible inventory in YAML format from all hosts
    # and tags in the database
    # @return       the Ansible YAML inventory as a string
    def getInventoryYaml(self):
        return yaml.dump(self.getInventory(), default_flow_style=False)

    # Gets a tag in the database
    # @param name       The name of the tag to get
    # @return           The Tag object, or None if the tag does
    #                   not exist.
    def getTag(self, name):
        cursor = self._conn.cursor()
        cursor.execute('''
            SELECT
                TagId,
                TagName,
                TagGroup,
                Description
            FROM Tag
            WHERE TagName = %s''',
            [name])

        row = cursor.fetchone()
        if row == None:
            return None

        tag = Tag(row[0], row[1], row[2], row[3], 
                self)
        cursor.fetchall()
        cursor.close()
        return tag

    # Gets all the tag groups in the database and which tags
    # belong to them.
    # @return   A list of (group, tagnames) tuples where tagnames
    #           is a comma separated list of tags in the group.
    def getTagGroups(self):
        groups = list()

        cursor = self._conn.cursor()
        cursor.execute("SELECT * FROM TagByGroup ORDER BY TagGroup ASC")
        for (groupName, tags) in cursor:
            group = groupName if groupName != None else 'ungrouped'
            groups.append( (group, tags) )
        cursor.close()

        return groups

    # Gets all the tags in the database
    # @param groupSort=False    If true, sort the tags first by
    #                           group name and then by tag name.
    #                           Otherwise just sort by tag name.
    # @return   An array containing all the tags in the database
    def getTags(self, groupSort=False):
        tags = list()

        # Build statement
        stmt = '''
            SELECT
                TagId,
                TagName,
                TagGroup,
                Description
            FROM Tag '''
        if groupSort == True:
            stmt += 'ORDER BY TagGroup ASC, TagName ASC'
        else:
            stmt += 'ORDER BY TagName ASC'

        # Fetch data
        cursor = self._conn.cursor()
        cursor.execute(stmt)
        for (tagId, name, group, description) in cursor:
            tag = Tag(tagId, name, group, description, self)
            tags.append(tag)
        cursor.close()

        return tags

    # Gets a dictionary of all the tags in the database broken down into
    # groups. The tag group will be the top level key in the dictionary and
    # each key's value will a list containing all the tags in that group.
    # @return   A dictionary containing all the tags in the database
    def getTagsByGroup(self):
        tags = {}

        # Build statement
        stmt = '''
            SELECT 
                TagId,
                TagName,
                TagGroup,
                Description
            FROM Tag '''
        if groupSort == True:
            stmt += 'ORDER BY TagGroup ASC, TagName ASC'
        else:
            stmt += 'ORDER BY TagName ASC'

        # Fetch data
        cursor = self._conn.cursor()
        cursor.execute(stmt)
        for (tagId, name, group, description) in cursor:
            tag = Tag(tagId, name, group, description, self)

            if group == None:
                group = 'ungrouped'
            if group not in tags:
                tags[group] = list()

            tags[group].append(tag)
        cursor.close()

        return tags

    # Gets the libIsidore version
    # @return       The libIsidore version
    def getVersion(self):
        return self._version

# An individual host
class Host:

    _hostId = None
    _hostname = None
    _commissionDate = None
    _decommissionDate = None
    _description = None
    _isidore = None

    # Creates a new Host object
    # @param hostId             The Isidore databases's internal ID for the host
    # @param hostname           The hostname for the host
    # @param commissionDate     The host's commission date
    # @param decommissionDate   The host's decommission date
    # @param description        The host's description
    # @param isidore            The Isidore object the host belongs to
    def __init__(self, hostId, hostname, commissionDate, decommissionDate, description, isidore=None):
        self._hostId = hostId
        self._hostname = hostname
        self._commissionDate = commissionDate
        self._decommissionDate = decommissionDate
        self._description = description
        self._isidore = isidore

    # Assigns a tag to this host
    # @param tag        The tag object to assign
    def addTag(self, tag):
        cursor = self._isidore._conn.cursor()
        stmt = "INSERT INTO HostHasTag (HostID, TagID) VALUES (%s, %s)"
        cursor.execute(stmt, [ self._hostId, tag.getTagId() ])
        self._isidore._conn.commit()
        cursor.close()

    # Appends an item to a list variable
    # @param path       The path of the list to append to.
    # @param value      The value to append to the list. This can
    #                   be of any type that is JSON serializable.
    def appendVar(self, path, value):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Host
            SET Variables =
                JSON_ARRAY_APPEND(
                    (SELECT Variables FROM Host WHERE HostId = %s),
                    %s,
                    JSON_EXTRACT(%s, '$')
                )
            WHERE HostID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [
            self._hostId,
            path,
            json.dumps(value),
            self._hostId])
        self._isidore._conn.commit()
        cursor.close()

    # Deletes this host from the database. The host object should
    # not be referenced after this method is called.
    def delete(self):
        # Delete the host
        cursor = self._isidore._conn.cursor()
        stmt = "DELETE FROM Host WHERE HostID = %s"
        cursor.execute(stmt, [ self._hostId ])
        self._isidore._conn.commit()
        cursor.close()

        # Blank out all the fields in case the object is
        # referenced again.
        self._hostId = None
        self._hostname = None
        self._commissionDate = None
        self._decommissionDate = None
        self._description = None
        self._isidore = None

    # Gets the host's internal ID in the Isidore database.
    # @return   The hosts's ID
    def getHostId(self):
        return self._hostId

    # Gets the host's hostname
    # @return   The hostname
    def getHostname(self):
        return self._hostname

    # Gets the host's commission date as a DateTime
    # @return   The commission date
    def getCommissionDate(self):
        return self._commissionDate

    # Gets the host's decommission date as a DateTime
    # @return   The decommission date, or None if the host is commissioned.
    def getDecommissionDate(self):
        return self._decommissionDate

    # Gets a dictionary containing all the details about this host
    # @return   A dictionary containing the details
    def getDetails(self):
        det = {}
        det[self._hostname] = {}
        det[self._hostname]['vars'] = self.getVar()
        isivar = {}

        # Host Attributes
        isivar['commissioned'] = str(self._commissionDate) \
                if self._commissionDate is not None else None
        isivar['decommissioned'] = str(self._decommissionDate) \
                if self._decommissionDate is not None else None
        isivar['description'] = self._description

        # Tags
        isivar['tags'] = {}
        for tag in self.getTags(True):
            group = tag.getGroup() if tag.getGroup() is not None else 'ungrouped'

            if group not in isivar['tags']:
                isivar['tags'][group] = list()

            isivar['tags'][group].append(tag.getName())

        det[self._hostname]['vars']['isidore'] = isivar
        return det

    # Gets the host's description
    # @return   The host's description
    def getDescription(self):
        return self._description

    # Gets all the tags assigned to this host
    # @param groupSort=False    If true, sort the tags first by
    #                           group name and then by tag name.
    #                           Otherwise just sort by tag name.
    # @return   An array containing all the tags assigned to this
    #           host
    def getTags(self, groupSort=False):
        tags = list()

        stmt = '''\
            SELECT
                Tag.TagID,
                TagName,
                TagGroup,
                Description
            FROM Tag
            INNER JOIN HostHasTag
                ON Tag.TagID = HostHasTag.TagID
            WHERE HostID = %s
            '''

        if groupSort == True:
            stmt += 'ORDER BY TagGroup ASC, TagName ASC'
        else:
            stmt += 'ORDER BY TagName ASC'

        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [self._hostId])
        for (tagId, name, group, description) in cursor:
            tag = Tag(tagId, name, group, description)
            tags.append(tag)
        cursor.close()

        return tags

    # Gets a dictionary of variables assigned to this host,
    # optionally starting at a certian path. Paths should be
    # specified as path.to.object. If no path is specified, all
    # variables will be included.
    # @param path       The path to start at. If not specifed,
    #                   defaults to the root of the JSON tree.
    # @return   A dictionary containing all the variables starting
    #           at path.
    def getVar(self, path='$'):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Select the JSON
        stmt = 'SELECT JSON_EXTRACT(Variables, %s) \
                FROM Host WHERE HostID = %s'
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, self._hostId])
        row = cursor.fetchone()
        cursor.close()

        # Decode the JSON and return
        if row[0] == None:
            return None
        else:
            return json.loads(row[0])

    # Removes a tag from this host
    # @param tag        The tag object to remove
    def removeTag(self, tag):
        cursor = self._isidore._conn.cursor()
        stmt = '''\
            DELETE FROM HostHasTag
            WHERE 
                HostID = %s AND
                TagId = %s
            '''
        cursor.execute(stmt, [ self._hostId, tag.getTagId() ])
        self._isidore._conn.commit()
        cursor.close()

    # Sets the host's commission date
    # @param date       The commission date
    def setCommissionDate(self, date):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Host SET CommissionDate = %s WHERE HostID = %s"
        cursor.execute(stmt, [ date, self._hostId ])
        self._isidore._conn.commit()
        cursor.close()
        self._commissionDate = date

    # Sets the host's decommission date
    # @param date       The decommission date, or None to clear the
    #                   decommission date.
    def setDecommissionDate(self, date):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Host SET DecommissionDate = %s WHERE HostID = %s"
        cursor.execute(stmt, [ date, self._hostId ])
        self._isidore._conn.commit()
        cursor.close()
        self._decommissionDate = date

    # Sets the host's description
    # @param description    The host's description
    def setDescription(self, description):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Host SET Description = %s WHERE HostID = %s"
        cursor.execute(stmt, [ description, self._hostId ])
        self._isidore._conn.commit()
        cursor.close()
        self._description = description

    # Sets the host's hostname
    # @param hostname       The hostname
    def setHostname(self, hostname):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Host SET Hostname = %s WHERE HostID = %s"
        cursor.execute(stmt, [ hostname, self._hostId ])
        self._isidore._conn.commit()
        cursor.close()
        self._hostname = hostname

    # Sets a variable to a specified value.
    # @param path       The path of the variable to set. It will
    #                   be created if it does not exist. If it is
    #                   not specifed or is $, the entire variable
    #                   tree will be overwritten with the
    #                   specified value.
    # @param value      The value to set the variable to. This can
    #                   be of any type that is JSON serializable.
    def setVar(self, path, value):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Host
            SET Variables =
                JSON_SET(
                    Variables,
                    %s,
                    JSON_EXTRACT(%s, '$')
                )
            WHERE HostID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, json.dumps(value), self._hostId])
        self._isidore._conn.commit()
        cursor.close()

    # Unsets a variable.
    # @param path       The path of the variable to unset.
    def unsetVar(self, path):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Host
            SET Variables =
                JSON_REMOVE(
                    Variables,
                    %s
                )
            WHERE HostID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, self._hostId])
        self._isidore._conn.commit()
        cursor.close()

# An individual tag
class Tag:
    
    _tagId = None
    _name = None
    _group = None
    _description = None
    _isidore = None

    # Creates a new Tag Object
    # @param tagID              The Isidore databases's internal ID for the tag
    # @param name               The name of the tag
    # @param group              The group for the tag
    # @param description        The tag's description
    # @param isidore            The Isidore object the tag belongs to
    def __init__(self, tagId, name, group, description,
            isidore=None):
        self._tagId = tagId
        self._name = name
        self._group = group
        self._description = description
        self._isidore = isidore

    # Appends an item to a list variable
    # @param path       The path of the list to append to.
    # @param value      The value to append to the list. This can
    #                   be of any type that is JSON serializable.
    def appendVar(self, path, value):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Tag
            SET Variables =
                JSON_ARRAY_APPEND(
                    (SELECT Variables FROM Tag WHERE TagID = %s),
                    %s,
                    JSON_EXTRACT(%s, '$')
                )
            WHERE TagID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [
            self._tagId,
            path,
            json.dumps(value),
            self._tagId])
        self._isidore._conn.commit()
        cursor.close()

    # Deletes this tag from the database. The tag object should
    # not be referenced after this method is called.
    def delete(self):
        # Delete the tag
        cursor = self._isidore._conn.cursor()
        stmt = "DELETE FROM Tag WHERE TagID = %s"
        cursor.execute(stmt, [ self._tagId ])
        self._isidore._conn.commit()
        cursor.close()

        # Blank out all the fields in case the object is
        # referenced again.
        self._tagId = None
        self._name = None
        self._group = None
        self._description = None
        self._isidore = None

    # Gets the tag's internal ID in the Isidore database.
    # @return   The tag's ID
    def getTagId(self):
        return self._tagId

    # Gets the tag's name
    # @return   The name
    def getName(self):
        return self._name

    # Gets the tag's description
    # @return   The tag's description
    def getDescription(self):
        return self._description

    # Gets a dictionary containing all the details about this tag
    # @return   A dictionary containing the details
    def getDetails(self):
        det = {}
        det[self._name] = {}
        det[self._name]['vars'] = self.getVar()
        isivar = {}
        hosts = list()

        # Tag Attributes
        isivar['description'] = self._description
        isivar['group'] = self._group

        # Hosts
        for host in self.getHosts():
            hosts.append(host.getHostname())

        det[self._name]['vars']['isidore_tag_'+self._name] = isivar
        det[self._name]['hosts'] = hosts
        return det

    def getGroup(self):
        return self._group

    # Gets all the commissioned hosts assigned to this tag
    # @return   An array containing all the commissioned hosts
    #           assigned to this tag
    def getHosts(self):
        hosts = list()

        stmt = '''\
            SELECT
                Host.HostID,
                Hostname,
                CommissionDate,
                DecommissionDate,
                Description
            FROM Host
            INNER JOIN HostHasTag
                ON Host.HostID = HostHasTag.HostID
            WHERE
                TagID = %s AND
                DecommissionDate IS NULL
            ORDER BY Hostname ASC
            '''

        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [self._tagId])
        for (hostId, hostname, commissionDate, decommissionDate, description) in cursor:
            host = Host(hostId, hostname, commissionDate,
                    decommissionDate, description, self)
            hosts.append(host)
        cursor.close()

        return hosts

    # Gets a dictionary of variables assigned to this tag,
    # optionally starting at a certian path. Paths should be
    # specified as path.to.object. If no path is specified, all
    # variables will be included.
    # @param path       The path to start at. If not specifed,
    #                   defaults to the root of the JSON tree.
    # @return   A dictionary containing all the variables starting
    #           at path.
    def getVar(self, path='$'):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Select the JSON
        stmt = 'SELECT JSON_EXTRACT(Variables, %s) \
                FROM Tag WHERE TagID = %s'
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, self._tagId])
        row = cursor.fetchone()
        cursor.close()

        # Decode the JSON and return
        if row[0] == None:
            return None
        else:
            return json.loads(row[0])

    # Sets the tag's description
    # @param description    The tag's description
    def setDescription(self, description):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Tag SET Description = %s WHERE TagID = %s"
        cursor.execute(stmt, [ description, self._tagId ])
        self._isidore._conn.commit()
        cursor.close()
        self._description = description

    # Sets the tag's group
    # @param group          The group for the tag, or None to remove it from
    #                       its current group.
    def setGroup(self, group):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Tag SET TagGroup = %s WHERE TagID = %s"
        cursor.execute(stmt, [ group, self._tagId ])
        self._isidore._conn.commit()
        cursor.close()
        self._group = group

    # Set the tag's name
    # @param name           The name
    def setName(self, name):
        cursor = self._isidore._conn.cursor()
        stmt = "UPDATE Tag SET TagName = %s WHERE TagID = %s"
        cursor.execute(stmt, [ name, self._tagId ])
        self._isidore._conn.commit()
        cursor.close()
        self._name = name

    # Sets a variable to a specified value.
    # @param path       The path of the variable to set. It will
    #                   be created if it does not exist. If it is
    #                   not specifed or is $, the entire variable
    #                   tree will be overwritten with the
    #                   specified value.
    # @param value      The value to set the variable to. This can
    #                   be of any type that is JSON serializable.
    def setVar(self, path, value):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Tag
            SET Variables =
                JSON_SET(
                    Variables,
                    %s,
                    JSON_EXTRACT(%s, '$')
                )
            WHERE TagID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, json.dumps(value), self._tagId])
        self._isidore._conn.commit()
        cursor.close()

    # Unsets a variable.
    # @param path       The path of the variable to unset.
    def unsetVar(self, path):
        # Ensure the path starts with $
        if path[0] != '$':
            path = '$.' + path

        # Set the variable
        stmt = '''
            UPDATE Tag
            SET Variables =
                JSON_REMOVE(
                    Variables,
                    %s
                )
            WHERE TagID = %s'''
        cursor = self._isidore._conn.cursor()
        cursor.execute(stmt, [path, self._tagId])
        self._isidore._conn.commit()
        cursor.close()

